/*
 * NCSA, University of Illinois Urbana-Champaign, 2004
 * Stuart Levy, slevy@ncsa.uiuc.edu
 */
/*
 * cc -o txspeed{,.c} -L/usr/X11R6/lib -lglut -lGL -lX11
 * Measures opengl texture update speed
 */
#define GL_GLEXT_PROTOTYPES 1

#if __APPLE__
# include <glut.h>
# include <OpenGL/glext.h>
#else
# include <X11/Xlib.h>
# include <GL/glut.h>
# include <GL/glext.h>
#endif

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <time.h>
#include <sys/time.h>
#include <signal.h>

#ifndef MAXTX
# define MAXTX 1024
#endif

#ifndef TXROOM
# define TXROOM  16
#endif

unsigned char txbuf[ TXROOM * MAXTX * MAXTX * 4 ];
int width = MAXTX;
int height = MAXTX;

int rectx = 0;
int recty = 0;

int use_range = 0;
int use_intact = 0;

enum PixelType { RGB, RGBA, BGRA, ABGR,YCBCR };

enum PixelType ptype = RGB;

struct pixeltype {
   char *name;
   int internalfmt;
   int fmt;
   int type;
  int size;
} ptypes[] = {
  { "rgb", GL_RGB,   GL_RGB,  GL_UNSIGNED_BYTE, sizeof(char) * 3 },
  { "rgba", GL_RGBA,  GL_RGBA, GL_UNSIGNED_BYTE, sizeof(char) * 4 },
  { "bgra", GL_RGBA,  GL_BGRA, GL_UNSIGNED_INT_8_8_8_8_REV, sizeof(unsigned int) },
  { "abgr", GL_RGBA,  GL_ABGR_EXT, GL_UNSIGNED_BYTE, sizeof(char) * 4 },
#if __APPLE__
  { "ycbcr", GL_RGB,   GL_YCBCR_422_APPLE, GL_UNSIGNED_SHORT_8_8_REV_APPLE, sizeof(unsigned short) },
#endif
   { NULL }
};

int newtex = 1;

int justone = 0;
int justload = 0;

#define TXSLOTS  1024

int winx, winy;
int nxtx, nytx, ntx;
GLuint txs[TXSLOTS];

int use_bufs = 0;
GLuint bufs[TXSLOTS];

void tick( int start );  /* tick(1) to start, tick(0) to continue, tick(-1) to clean up */

void reshape( int xsize, int ysize )
{
    int i;

    winx = xsize;  winy = ysize;

    glMatrixMode( GL_PROJECTION );
    glLoadIdentity();
    glOrtho( 0, xsize, 0, ysize, -1, 1 );
    glViewport( 0, 0, xsize, ysize );
    glMatrixMode( GL_MODELVIEW );
    glLoadIdentity();

    glTexEnvf( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_DECAL );
    glTexParameterf( GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_MIN_FILTER, GL_NEAREST );
    glTexParameterf( GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_MAG_FILTER, GL_NEAREST );

    glEnable( GL_TEXTURE_RECTANGLE_ARB );	// was GL_TEXTURE_2D

    width = rectx ? rectx : MAXTX;
    height = recty ? recty : MAXTX;

    /* How many textures do we need to fill the screen? */
    nxtx = winx / width;  if(nxtx == 0) nxtx = 1;
    nytx = winy / height;  if(nytx == 0) nytx = 1;
    if(nxtx*nytx > TXSLOTS) nytx = TXSLOTS / nxtx;

    /* ensure that that many texture-objects exist */
    for(i = 0; i < nxtx*nytx; i++) {
	if(txs[i] == 0) {
	    glGenTextures( nxtx*nytx-i, &txs[i] );
	    if(use_bufs) glGenBuffers( nxtx*nytx-i, &bufs[i] );
	    newtex = 1;
	    break;
	}
    }

    for(i = 0; i < sizeof(txbuf); i+=4) {
	int v = i ^ (i/width) ^ (i/width/height);
	txbuf[i] = v;
	txbuf[i+1] = (v<<1) | (v>>7);
	txbuf[i+2] = (v<<3) | (v>>5);
	txbuf[i+3] = (v<<5) | (v>>3);
    }

    tick(1);
    printf("reshape %dx%d  %dx%d\n", xsize, ysize, nxtx, nytx);
    printf(" %-12s\t%-12s\t%s\n", "ms/frame", "Mpix/sec", "unixtime");
}

int quitnow = 0;

void idler()
{
    if(quitnow) {
	tick(-1);
	exit(1);
    }
    glutPostRedisplay();
}

void catch(int sig) {
    quitnow = 1;
}

int ntimes = 50;

void tick( int start )
{
    static struct timeval then;
    struct timeval now;
    static int count;
    double dt;

    if(start>0) {
	count = 0;
	return;
    }

    count++;

    if(count == 2) {
	gettimeofday( &then, NULL );
    } else if(count == ntimes+2 || start<0) {
	gettimeofday( &now, NULL );
	dt = (now.tv_sec - then.tv_sec) + 1e-6*(now.tv_usec - then.tv_usec);
	if(count > 2)
	    printf("  %-12g\t%-12g\t%ld\n", 1000*dt/(count), 1e-6*width*height*(float)nxtx*nytx*(count)/dt, (long) now.tv_sec);
	else printf("tick done: count %d dt %g\n", count, dt);
	count = 0;
    }
}

void kb( unsigned char key, int x, int y )
{
    tick( -1 );
    exit(0);
}

void redraw()
{
    int i, j, k, start, e;
    int rowlength = 0;

    { static int once = 1;
      if(once) {
	GLint mts;
	glGetIntegerv( GL_MAX_TEXTURE_SIZE, &mts );
	printf("max texture size %d\n", (int)mts);
	once = 0;
      }
    }

    glClearColor( .3, newtex, 0, 1 );
    glClear( GL_COLOR_BUFFER_BIT );

    glEnable( GL_TEXTURE_RECTANGLE_ARB );	// was GL_TEXTURE_2D

#ifdef __APPLE__
    if(newtex) {
	/* APPLE_texture_range: suggest that our texture area should be mapped for direct access by graphics card */
	if(use_range) {
	    glTexParameteri( GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_STORAGE_HINT_APPLE, GL_STORAGE_SHARED_APPLE );
	    glTextureRangeAPPLE( GL_TEXTURE_RECTANGLE_ARB, sizeof(txbuf), txbuf );
	}

	/* Promise OpenGL we'll leave our texture buffer intact -- avoid copy */
	if(use_intact)
	    glPixelStorei( GL_UNPACK_CLIENT_STORAGE_APPLE, GL_TRUE );
    }
#endif

    /* Load textures and draw */
    for(i = k = 0; i < nytx; i++) {
	for(j = 0; j < nxtx; j++, k++) {

	    int x0 = j*width;
	    int x1 = (j+1)*width;
	    int y0 = i*height;
	    int y1 = (i+1)*height;

	    if(x1 > winx) x1 = winx;
	    if(y1 > winy) y1 = winy;

	    if(!justone) {
	      glBindTexture( GL_TEXTURE_RECTANGLE_ARB, txs[k] );	// was GL_TEXTURE_2D
	      if(use_bufs) glBindBuffer(GL_PIXEL_UNPACK_BUFFER_ARB, bufs[k]);
	    }
	    if(rowlength == 0)
		start = (int) ( (((TXROOM-1)*MAXTX*MAXTX)*4) * drand48() );
	    else
		start = (int) ((MAXTX*MAXTX*4) * drand48());

	    while(rectx*recty>0 && start > 0 && start > sizeof(txbuf) - rectx*recty*4)
		start /= 2;	/* hack -- don't offer beyond end of txbuf */

	    if(newtex) {

	      if(use_bufs) {
		glBufferData(GL_PIXEL_UNPACK_BUFFER_ARB,width * height * ptypes[ptype].size,&txbuf[(start / 16) * 16],GL_STATIC_DRAW);
	      }
	      
	      glPixelStorei( GL_UNPACK_ROW_LENGTH, rowlength );
	      /* glEnable( GL_TEXTURE_2D ); */
	      glTexEnvf( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_DECAL );
	      glTexParameterf( GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_MIN_FILTER, GL_NEAREST );
	      glTexParameterf( GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_MAG_FILTER, GL_NEAREST );
	      
	      glTexImage2D( GL_TEXTURE_RECTANGLE_ARB, 0, ptypes[ptype].internalfmt, width, height, 0,
			    ptypes[ptype].fmt, ptypes[ptype].type, use_bufs ? 0 : &txbuf[(start / 16) * 16] );
	    } else {
	      if(use_bufs) {
		glBufferSubData(GL_PIXEL_UNPACK_BUFFER_ARB,0,width * height * ptypes[ptype].size,&txbuf[(start / 16) * 16]);
	      }
	      glTexSubImage2D( GL_TEXTURE_RECTANGLE_ARB, 0,  0,0,  width, height,
			       ptypes[ptype].fmt, ptypes[ptype].type, use_bufs ? 0 : &txbuf[(start / 16) * 16] );
	    }

	    if(!justload) {
	    glBegin( GL_QUADS );

	    glTexCoord2f( 0, 0 );
	    glVertex3i( x0, y0, 0 );

	    glTexCoord2f( 0, height );
	    glVertex3i( x0, y1, 0 );

	    glTexCoord2f( width, height );
	    glVertex3i( x1, y1, 0 );

	    glTexCoord2f( width, 0 );
	    glVertex3i( x1, y0, 0 );
	    glEnd();
	    }

	    e = glGetError();
	    if(e != 0)
		fprintf(stderr, "E%d ", e);
	}
    }

    glutSwapBuffers();

    newtex = 0;

    tick(0);

}

int main(int argc, char *argv[])
{
    char *prog = argv[0];

    signal(SIGINT, catch);
    justone = (getenv("JUSTONE") != NULL);
    justload = (getenv("JUSTLOAD") != NULL);

    setlinebuf(stdout);

    glutInit( &argc, argv );

    while(argc>1) {
	enum PixelType pt;
	int found = 0;

	if(argc>2 && 0==strcmp(argv[1], "-W")) {
	    if(1==sscanf(argv[2], "%d%*c%d", &rectx, &recty))
		recty = rectx;
	    argc -= 2;
	    argv += 2;
	    continue;
	}

#if __APPLE__
	if(0==strcmp(argv[1], "-range")) {
	    use_range = 1;
	    found = 1;
	} else if(0==strcmp(argv[1], "-intact")) {
	    use_intact = 1;
	    found = 1;
	}
#endif
	if(0 == strcmp(argv[1],"-pbo")) {
	  use_bufs = 1;
	  argc -=1;
	  argv += 2;
	  continue;
	}

	for(pt = 0; ptypes[pt].name!=NULL && !found; pt++) {
	    if(argv[1][0] == '-' && 0==strcasecmp(&argv[1][1], ptypes[pt].name)) {
		ptype = pt;
		found = 1;
	    }
	}
	if(!found) {
#if __APPLE__
	    fprintf(stderr, "Usage: %s [-W txwidth,txheight]  [-rgb | -rgba | -bgra | -abgr | -ycbcr]  [-range] [-intact]\n\
Do texture performance test with rectangular textures, and various pixel formats.\n\
 -range: Use Apple texture-range extension (maps textures for direct GPU access)\n\
 -intact: Promise to leave texturebuffers intact (GL_UNPACK_CLIENT_STORAGE_APPLE) (avoids extra copy)\n\
Default -W %d,%d -rgb\n", prog, MAXTX,MAXTX);
#else
	    fprintf(stderr, "Usage: %s [-W txwidth,txheight]  [-rgb | -rgba | -bgra | -abgr]\n\
Do texture performance test with rectangular textures, and various pixel formats.\n\
Default -W %d,%d -rgb\n", prog, MAXTX,MAXTX);
#endif
	    exit(1);
	}
	argc--, argv++;
    }

    glutInitDisplayMode( GLUT_RGB | GLUT_DOUBLE );

    glutCreateWindow( "eep" );
    glutFullScreen();

    glutIdleFunc( idler );
    glutReshapeFunc( reshape );
    glutDisplayFunc( redraw );
    glutKeyboardFunc( kb );

    glutMainLoop();
    return 0;
}
